'''
Function:
    视频转字符画
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import numpy as np
from ..base import BaseBeautifier
from PIL import Image, ImageFont, ImageDraw


'''视频转字符画'''
class CharacterizeBeautifier(BaseBeautifier):
    CHARS = "$@B%8&WM#*oahkbdpqwmZO0QLCJUYXzcvunxrjft/\|()1{}[]?-_+~<>i!lI;:,\"^`'. "
    def __init__(self, **kwargs):
        super(CharacterizeBeautifier, self).__init__(**kwargs)
    '''迭代图片'''
    def iterimage(self, image):
        image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        # 每个字符的大小
        font = ImageFont.load_default().font
        font_w, font_h = font.getsize(self.CHARS[1])
        # 输入图像变成可整除字符大小
        image = image.resize((font_w * (image.width // font_w), font_h * (image.height // font_h)), Image.NEAREST)
        # 原始大小
        h_ori, w_ori = image.height, image.width
        # resize
        image = image.resize((w_ori // font_w, h_ori // font_h), Image.NEAREST)
        h, w = image.height, image.width
        # 图像RGB转字符
        txts, colors = '', []
        for i in range(h):
            for j in range(w):
                pixel = image.getpixel((j, i))
                colors.append(pixel[:3])
                txts += self.rgb2char(*pixel)
        image = Image.new('RGB', (w_ori, h_ori), (255, 255, 255))
        draw = ImageDraw.Draw(image)
        x = y = 0
        for j in range(len(txts)):
            if x == w_ori: x, y = 0, y + font_h
            draw.text((x, y), txts[j], font=font, fill=colors[j])
            x += font_w
        return cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
    '''RGB转字符'''
    def rgb2char(self, r, g, b, alpha=256):
        if alpha == 0: return ''
        gray = int(0.2126 * r + 0.7152 * g + 0.0722 * b)
        return self.CHARS[gray % len(self.CHARS)]


'''
Function:
    图像卡通化
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import numpy as np
from ..base import BaseBeautifier


'''图像卡通化'''
class CartooniseBeautifier(BaseBeautifier):
    def __init__(self, mode='rgb', **kwargs):
        super(CartooniseBeautifier, self).__init__(**kwargs)
        assert mode in ['rgb', 'hsv']
        self.mode = mode
    '''迭代图片'''
    def iterimage(self, image):
        if self.mode == 'rgb':
            return self.processinrgb(image)
        elif self.mode == 'hsv':
            return self.processinhsv(image)
    '''在RGB空间操作'''
    def processinrgb(self, image):
        # Step1: 利用双边滤波器对原图像进行保边去噪处理
        # --下采样
        image_bilateral = image
        for _ in range(2):
            image_bilateral = cv2.pyrDown(image_bilateral)
        # --进行多次的双边滤波
        for _ in range(7):
            image_bilateral = cv2.bilateralFilter(image_bilateral, d=9, sigmaColor=9, sigmaSpace=7)
        # --上采样
        for _ in range(2):
            image_bilateral = cv2.pyrUp(image_bilateral)
        # Step2: 将步骤一中获得的图像灰度化后，使用中值滤波器去噪
        image_gray = cv2.cvtColor(image_bilateral, cv2.COLOR_RGB2GRAY)
        image_median = cv2.medianBlur(image_gray, 7)
        # Step3: 对步骤二中获得的图像使用自适应阈值从而获得原图像的轮廓
        image_edge = cv2.adaptiveThreshold(image_median, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, blockSize=9, C=2)
        image_edge = cv2.cvtColor(image_edge, cv2.COLOR_GRAY2RGB)
        # Step4: 将步骤一中获得的图像与步骤三中获得的图像轮廓合并即可实现将照片变为卡通图片的效果了
        image_cartoon = cv2.bitwise_and(image_bilateral, image_edge)
        # 返回
        return image_cartoon
    '''在HSV空间操作'''
    def processinhsv(self, image):
        # Step1: 图像BGR空间转HSV空间, 在HSV空间进行直方图均衡化, 中值滤波和形态学变换
        image_hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        h, s, v = cv2.split(image_hsv)
        # --直方图均衡化
        v = cv2.equalizeHist(v)
        image_hsv = cv2.merge((h, s, v))
        # --中值滤波
        image_hsv = cv2.medianBlur(image_hsv, 7)
        # --形态学变换-开/闭运算
        kernel = np.ones((5, 5), np.uint8)
        image_hsv = cv2.morphologyEx(image_hsv, cv2.MORPH_CLOSE, kernel, iterations=2)
        # --中值滤波
        image_hsv = cv2.medianBlur(image_hsv, 7)
        # Step2: 对步骤一中获得的图像使用自适应阈值从而获得原图像的轮廓
        image_mask = cv2.cvtColor(image_hsv, cv2.COLOR_HSV2BGR)
        image_mask = cv2.cvtColor(image_mask, cv2.COLOR_RGB2GRAY)
        image_mask = cv2.medianBlur(image_mask, 7)
        image_edge = cv2.adaptiveThreshold(image_mask, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, blockSize=9, C=2)
        image_edge = cv2.cvtColor(image_edge, cv2.COLOR_GRAY2RGB)
        # Step3: 将步骤二中获得的图像轮廓与原图像合并即可实现将照片变为卡通图片的效果了
        image_cartoon = cv2.bitwise_and(image, image_edge)
        # 返回
        return image_cartoon


'''
Function:
    图像抖音特效画
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import copy
import numpy as np
from PIL import Image
from ..base import BaseBeautifier


'''图像抖音特效画'''
class DouyinEffectBeautifier(BaseBeautifier):
    def __init__(self, **kwargs):
        super(DouyinEffectBeautifier, self).__init__(**kwargs)
    '''迭代图片'''
    def iterimage(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGBA)
        # 提取R
        image_arr_r = copy.deepcopy(image)
        image_arr_r[:, :, 1:3] = 0
        # 提取GB
        image_arr_gb = copy.deepcopy(image)
        image_arr_gb[:, :, 0] = 0
        # 创建画布把图片错开放
        image_r = Image.fromarray(image_arr_r).convert('RGBA')
        image_gb = Image.fromarray(image_arr_gb).convert('RGBA')
        canvas_r = Image.new('RGB', (image.shape[1], image.shape[0]), color=(0, 0, 0))
        canvas_gb = Image.new('RGB', (image.shape[1], image.shape[0]), color=(0, 0, 0))
        canvas_r.paste(image_r, (6, 6), image_r)
        canvas_gb.paste(image_gb, (0, 0), image_gb)
        output_image = np.array(canvas_gb) + np.array(canvas_r)
        output_image = cv2.cvtColor(output_image, cv2.COLOR_RGB2BGR)
        # 返回结果
        return output_image


'''initialize'''
from .glitch import GlitchBeautifier
from .cartoonise import CartooniseBeautifier
from .cartoongan import CartoonGanBeautifier
from .beziercurve import BezierCurveBeautifier
from .photomosaic import PhotomosaicBeautifier
from .characterize import CharacterizeBeautifier
from .douyineffect import DouyinEffectBeautifier
from .pencildrawing import PencilDrawingBeautifier
from .cartoonizeface import CartoonizeFaceBeautifier
from .fastneuralstyletransfer import FastNeuralStyleTransferBeautifier
from .geneticfitting import GeneticFittingCircleBeautifier, GeneticFittingPolygonBeautifier


'''
Function:
    IO相关的工具函数
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import cv2
from tqdm import tqdm


'''检查文件是否存在'''
def checkdir(dirname):
    if os.path.exists(dirname): return True
    os.mkdir(dirname)
    return False


'''将图片转为视频并保存'''
def Images2VideoAndSave(images, savedir='outputs', savename='output', fps=25, ext='avi', logger_handle=None):
    checkdir(savedir)
    savepath = os.path.join(savedir, savename + f'.{ext}')
    fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
    video_writer = cv2.VideoWriter(savepath, fourcc, fps, (images[0].shape[1], images[0].shape[0]))
    pbar = tqdm(images)
    for image in pbar:
        pbar.set_description(f'Writing image to {savepath}')
        video_writer.write(image)
    if logger_handle is not None: logger_handle.info(f'Video is saved into {savepath}')


'''保存图片'''
def SaveImage(image, savedir='outputs', savename='output', ext='png', logger_handle=None):
    checkdir(savedir)
    savepath = os.path.join(savedir, savename + f'.{ext}')
    cv2.imwrite(savepath, image)
    if logger_handle is not None: logger_handle.info(f'Image is saved into {savepath}')


'''读取视频'''
def ReadVideo(videopath):
    capture, images = cv2.VideoCapture(videopath), []
    while capture.isOpened():
        ret, frame = capture.read()
        if not ret: break
        images.append(frame)
    capture.release()
    return images


'''title'''
__title__ = 'pydrawing'
'''description'''
__description__ = 'Pydrawing: Beautify your image or video.'
'''url'''
__url__ = 'https://github.com/CharlesPikachu/pydrawing'
'''version'''
__version__ = '0.1.3'
'''author'''
__author__ = 'Zhenchao Jin'
'''email'''
__email__ = 'charlesblwx@gmail.com'
'''license'''
__license__ = 'MIT License'
'''copyright'''
__copyright__ = 'Copyright 2021 Zhenchao Jin'


'''
Function:
    Beautifier基类
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import cv2
import subprocess
from tqdm import tqdm
from ...utils import Images2VideoAndSave, SaveImage, ReadVideo


'''Beautifier基类'''
class BaseBeautifier():
    def __init__(self, savedir='outputs', savename='output', **kwargs):
        self.savename, self.savedir = savename, savedir
        self.merge_audio, self.tmp_audio_path = False, 'cache.mp3'
        for key, value in kwargs.items(): setattr(self, key, value)
    '''处理文件'''
    def process(self, filepath, images=None):
        assert images is None or filepath is None, 'please input filepath or images rather than both'
        # 图片/视频处理
        if images is None:
            # --图片
            if filepath.split('.')[-1].lower() in ['jpg', 'jpeg', 'png']:
                images = [cv2.imread(filepath)]
            # --视频
            elif filepath.split('.')[-1].lower() in ['mp4', 'avi']:
                images = ReadVideo(filepath)
                if self.merge_audio:
                    p = subprocess.Popen(f'ffmpeg -i {filepath} -f mp3 {self.tmp_audio_path}')
                    while True:
                        if subprocess.Popen.poll(p) is not None: break
            # --不支持的数据格式
            else:
                raise RuntimeError('Unsupport file type %s...' % filepath.split('.')[-1])
        outputs, pbar = [], tqdm(images)
        for image in pbar:
            pbar.set_description('Process image')
            outputs.append(self.iterimage(image))
        if len(outputs) > 1:
            fps, ext = 25, 'avi'
            if hasattr(self, 'fps'): fps = self.fps
            if hasattr(self, 'ext'): ext = self.ext
            Images2VideoAndSave(outputs, savedir=self.savedir, savename=self.savename, fps=fps, ext=ext, logger_handle=self.logger_handle)
        else:
            ext = 'png'
            if hasattr(self, 'ext'): ext = self.ext
            SaveImage(outputs[0], savedir=self.savedir, savename=self.savename, ext=ext, logger_handle=self.logger_handle)
        # 如果有音频, 则把音频merge进新视频
        if self.merge_audio:
            p = subprocess.Popen(f'ffmpeg -i {os.path.join(self.savedir, self.savename+f".{ext}")} -i {self.tmp_audio_path} -strict -2 -f mp4 {os.path.join(self.savedir, self.savename+"_audio.mp4")}')
            while True:
                if subprocess.Popen.poll(p) is not None: break
            os.remove(self.tmp_audio_path)
            self.logger_handle.info(f'Video with merged audio is saved into {os.path.join(self.savedir, self.savename+"_audio.mp4")}')
    '''迭代图片'''
    def iterimage(self, image):
        raise NotImplementedError('not to be implemented')


'''
Function:
    人脸检测
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import ssl
import math
import numpy as np
ssl._create_default_https_context = ssl._create_unverified_context


'''FaceDetector'''
class FaceDetector():
    def __init__(self, use_cuda, **kwargs):
        super(FaceDetector, self).__init__()
        try:
            import face_alignment
        except:
            raise RuntimeError('Please run "pip install face_alignment" to install "face_alignment"')
        device = 'cuda' if use_cuda else 'cpu'
        self.dlib_detector = face_alignment.FaceAlignment(face_alignment.LandmarksType._2D, device=device, face_detector='dlib')
    '''forward'''
    def __call__(self, image):
        # obtain landmarks
        preds = self.dlib_detector.get_landmarks(image)
        landmarks = None
        if preds is None:
            raise RuntimeError('no faces are detected')
        elif len(preds) == 1:
            landmarks = preds[0]
        else:
            areas = []
            for pred in preds:
                landmarks_top = np.min(pred[:, 1])
                landmarks_bottom = np.max(pred[:, 1])
                landmarks_left = np.min(pred[:, 0])
                landmarks_right = np.max(pred[:, 0])
                areas.append((landmarks_bottom - landmarks_top) * (landmarks_right - landmarks_left))
            max_face_index = np.argmax(areas)
            landmarks = preds[max_face_index]
        # rotate
        left_eye_corner = landmarks[36]
        right_eye_corner = landmarks[45]
        radian = np.arctan((left_eye_corner[1] - right_eye_corner[1]) / (left_eye_corner[0] - right_eye_corner[0]))
        height, width, _ = image.shape
        cos = math.cos(radian)
        sin = math.sin(radian)
        new_w = int(width * abs(cos) + height * abs(sin))
        new_h = int(width * abs(sin) + height * abs(cos))
        Tx = new_w // 2 - width // 2
        Ty = new_h // 2 - height // 2
        M = np.array([[cos, sin, (1 - cos) * width / 2. - sin * height / 2. + Tx], [-sin, cos, sin * width / 2. + (1 - cos) * height / 2. + Ty]])
        image_rotate = cv2.warpAffine(image, M, (new_w, new_h), borderValue=(255, 255, 255))
        landmarks = np.concatenate([landmarks, np.ones((landmarks.shape[0], 1))], axis=1)
        landmarks_rotate = np.dot(M, landmarks.T).T
        # return
        return self.crop(image_rotate, landmarks_rotate)
    '''crop'''
    def crop(self, image, landmarks):
        landmarks_top = np.min(landmarks[:, 1])
        landmarks_bottom = np.max(landmarks[:, 1])
        landmarks_left = np.min(landmarks[:, 0])
        landmarks_right = np.max(landmarks[:, 0])
        top = int(landmarks_top - 0.8 * (landmarks_bottom - landmarks_top))
        bottom = int(landmarks_bottom + 0.3 * (landmarks_bottom - landmarks_top))
        left = int(landmarks_left - 0.3 * (landmarks_right - landmarks_left))
        right = int(landmarks_right + 0.3 * (landmarks_right - landmarks_left))
        if bottom - top > right - left:
            left -= ((bottom - top) - (right - left)) // 2
            right = left + (bottom - top)
        else:
            top -= ((right - left) - (bottom - top)) // 2
            bottom = top + (right - left)
        image_crop = np.ones((bottom - top + 1, right - left + 1, 3), np.uint8) * 255
        h, w = image.shape[:2]
        left_white = max(0, -left)
        left = max(0, left)
        right = min(right, w-1)
        right_white = left_white + (right-left)
        top_white = max(0, -top)
        top = max(0, top)
        bottom = min(bottom, h-1)
        bottom_white = top_white + (bottom - top)
        image_crop[top_white:bottom_white+1, left_white:right_white+1] = image[top:bottom+1, left:right+1].copy()
        return image_crop


'''initialize'''
from .utils import Logger
from .beautifiers import (
    CartooniseBeautifier, PencilDrawingBeautifier, CartoonGanBeautifier, FastNeuralStyleTransferBeautifier,
    DouyinEffectBeautifier, CharacterizeBeautifier, PhotomosaicBeautifier, GlitchBeautifier, CartoonizeFaceBeautifier,
    GeneticFittingCircleBeautifier, GeneticFittingPolygonBeautifier, BezierCurveBeautifier
)


'''
Function:
    人脸分割
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import torch
import numpy as np
import torch.nn as nn
import torch.utils.model_zoo as model_zoo


'''config'''
SEGMENTOR_CFG = {
    'type': 'ce2p',
    'benchmark': True,
    'num_classes': -1,
    'align_corners': False,
    'is_multi_gpus': True,
    'distributed': {'is_on': True, 'backend': 'nccl'},
    'norm_cfg': {'type': 'batchnorm2d', 'opts': {}},
    'act_cfg': {'type': 'leakyrelu', 'opts': {'negative_slope': 0.01, 'inplace': True}},
    'backbone': {
        'type': 'resnet101',
        'series': 'resnet',
        'pretrained': False,
        'outstride': 16,
        'use_stem': True,
        'selected_indices': (0, 1, 2, 3),
    },
    'ppm': {
        'in_channels': 2048,
        'out_channels': 512,
        'pool_scales': [1, 2, 3, 6],
    },
    'epm': {
        'in_channels_list': [256, 512, 1024],
        'hidden_channels': 256,
        'out_channels': 2
    },
    'shortcut': {
        'in_channels': 256,
        'out_channels': 48,
    },
    'decoder':{ 
        'stage1': {
            'in_channels': 560,
            'out_channels': 512,
            'dropout': 0,
        },
        'stage2': {
            'in_channels': 1280,
            'out_channels': 512,
            'dropout': 0.1
        },
    },
}
SEGMENTOR_CFG.update(
    {
        'num_classes': 20,
        'backbone': {
            'type': 'resnet50',
            'series': 'resnet',
            'pretrained': True,
            'outstride': 8,
            'use_stem': True,
            'selected_indices': (0, 1, 2, 3),
        }
    }
)


'''FaceSegmentor'''
class FaceSegmentor(nn.Module):
    def __init__(self, **kwargs):
        super(FaceSegmentor, self).__init__()
        try:
            from ssseg.modules.models.segmentors.ce2p import CE2P
        except:
            raise RuntimeError('Please run "pip install sssegmentation" to install "ssseg"')
        self.ce2p = CE2P(SEGMENTOR_CFG, mode='TEST')
        self.ce2p.load_state_dict(model_zoo.load_url('https://github.com/SegmentationBLWX/modelstore/releases/download/ssseg_ce2p/ce2p_resnet50os8_lip_train.pth')['model'])
    '''forward'''
    def forward(self, x):
        return self.ce2p(x)
    '''preprocess'''
    def preprocess(self, image):
        # Resize
        output_size = (473, 473)
        if image.shape[0] > image.shape[1]:
            dsize = min(output_size), max(output_size)
        else:
            dsize = max(output_size), min(output_size)
        image = cv2.resize(image, dsize=dsize, interpolation=cv2.INTER_LINEAR)
        # Normalize
        mean, std = np.array([123.675, 116.28, 103.53]), np.array([58.395, 57.12, 57.375])
        image = image.astype(np.float32)
        mean = np.float64(mean.reshape(1, -1))
        stdinv = 1 / np.float64(std.reshape(1, -1))
        cv2.cvtColor(image, cv2.COLOR_BGR2RGB, image)
        cv2.subtract(image, mean, image)
        cv2.multiply(image, stdinv, image)
        # ToTensor
        image = torch.from_numpy((image.transpose((2, 0, 1))).astype(np.float32))
        # Return
        return image.unsqueeze(0)
    '''get face mask'''
    def getfacemask(self, mask):
        output_mask = np.zeros(mask.shape[:2])
        face_idxs = [2, 13]
        for idx in face_idxs:
            output_mask[mask == idx] = 255
        return output_mask


'''
Function:
    利用贝塞尔曲线画画
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import re
import cv2
import turtle
import numpy as np
from bs4 import BeautifulSoup
from ..base import BaseBeautifier


'''利用贝塞尔曲线画画'''
class BezierCurveBeautifier(BaseBeautifier):
    def __init__(self, num_samples=15, width=600, height=600, num_colors=32, **kwargs):
        super(BezierCurveBeautifier, self).__init__(**kwargs)
        self.num_samples = num_samples
        self.width = width
        self.height = height
        self.num_colors = num_colors
        self.rootdir = os.path.split(os.path.abspath(__file__))[0]
    '''迭代图片'''
    def iterimage(self, image):
        data = image.reshape((-1, 3))
        data = np.float32(data)
        # 聚类迭代停止的模式(停止的条件, 迭代最大次数, 精度)
        criteria = (cv2.TERM_CRITERIA_EPS, 10, 1.0)
        # 数据, 分类数, 预设的分类标签, 迭代停止的模式, 重复试验kmeans算法次数, 初始类中心的选择方式
        compactness, labels, centers = cv2.kmeans(data, self.num_colors, None, criteria, 10, cv2.KMEANS_RANDOM_CENTERS)
        centers = np.uint8(centers)
        data_compress = centers[labels.flatten()]
        img_new = data_compress.reshape(image.shape)
        count = 0
        for center in centers:
            count += 1
            part = cv2.inRange(img_new, center, center)
            part = cv2.bitwise_not(part)
            cv2.imwrite('.tmp.bmp', part)
            os.system(f'{os.path.join(self.rootdir, "potrace.exe")} .tmp.bmp -s --flat')
            if count == 1:
                self.drawsvg('.tmp.svg', '#%02x%02x%02x' % (center[2], center[1], center[0]), True)
            else:
                self.drawsvg('.tmp.svg', '#%02x%02x%02x' % (center[2], center[1], center[0]), False)
        os.remove('.tmp.bmp')
        os.remove('.tmp.svg')
        turtle.done()
        return image
    '''一阶贝塞尔'''
    def FirstOrderBezier(self, p0, p1, t):
        assert (len(p0) == 2 and len(p1) == 2) or (len(p0) == 1 and len(p1) == 1)
        if len(p0) == 2 and len(p1) == 2:
            return p0[0] * (1 - t) + p1[0] * t, p0[1] * (1 - t) + p1[1] * t
        else:
            return p0 * (1 - t) + p1 * t
    '''二阶贝塞尔'''
    def SecondOrderBezier(self, p0, p1, p2):
        turtle.goto(p0)
        turtle.pendown()
        for t in range(0, self.num_samples+1):
            p = self.FirstOrderBezier(self.FirstOrderBezier(p0, p1, t/self.num_samples), self.FirstOrderBezier(p1, p2, t/self.num_samples), t/self.num_samples)
            turtle.goto(p)
        turtle.penup()
    '''三阶贝塞尔'''
    def ThirdOrderBezier(self, p0, p1, p2, p3):
        p0 = -self.width / 2 + p0[0], self.height / 2 - p0[1]
        p1 = -self.width / 2 + p1[0], self.height / 2 - p1[1]
        p2 = -self.width / 2 + p2[0], self.height / 2 - p2[1]
        p3 = -self.width / 2 + p3[0], self.height / 2 - p3[1]
        turtle.goto(p0)
        turtle.pendown()
        for t in range(0, self.num_samples+1):
            p = self.FirstOrderBezier(
                self.FirstOrderBezier(self.FirstOrderBezier(p0, p1, t/self.num_samples), self.FirstOrderBezier(p1, p2, t/self.num_samples), t/self.num_samples), 
                self.FirstOrderBezier(self.FirstOrderBezier(p1, p2, t/self.num_samples), self.FirstOrderBezier(p2, p3, t/self.num_samples), t/self.num_samples), 
                t/self.num_samples
            )
            turtle.goto(p)
        turtle.penup()
    '''画图(SVG)'''
    def drawsvg(self, filename, color, is_first=True, speed=1000):
        svgfile = open(filename, 'r')
        soup = BeautifulSoup(svgfile.read(), 'lxml')
        height, width = float(soup.svg.attrs['height'][:-2]), float(soup.svg.attrs['width'][:-2])
        scale = tuple(map(float, re.findall(r'scale\((.*?)\)', soup.g.attrs['transform'])[0].split(',')))
        scale = scale[0], -scale[1]
        if is_first:
            turtle.setup(height=height, width=width)
            turtle.setworldcoordinates(-width/2, 300, width-width/2, -height+300)
        turtle.tracer(100)
        turtle.pensize(1)
        turtle.speed(speed)
        turtle.penup()
        turtle.color(color)
        for path in soup.find_all('path'):
            attrs = path.attrs['d'].replace('\n', ' ')
            attrs = attrs.split(' ')
            attrs_yield = self.yieldattrs(attrs)
            endl = ''
            for attr in attrs_yield:
                if attr == 'M':
                    turtle.end_fill()
                    x, y = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.penup()
                    turtle.goto(-self.width/2+x, self.height/2-y)
                    turtle.pendown()
                    turtle.begin_fill()
                elif attr == 'm':
                    turtle.end_fill()
                    dx, dy = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.penup()
                    turtle.goto(turtle.xcor()+dx, turtle.ycor()-dy)
                    turtle.pendown()
                    turtle.begin_fill()
                elif attr == 'C':
                    p1 = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    p2 = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    p3 = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.penup()
                    p0 = turtle.xcor() + self.width / 2, self.height / 2 - turtle.ycor()
                    self.ThirdOrderBezier(p0, p1, p2, p3)
                    endl = attr
                elif attr == 'c':
                    turtle.penup()
                    p0 = turtle.xcor() + self.width / 2, self.height / 2 - turtle.ycor()
                    p1 = attrs_yield.__next__() * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    p2 = attrs_yield.__next__() * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    p3 = attrs_yield.__next__() * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    self.ThirdOrderBezier(p0, p1, p2, p3)
                    endl = attr
                elif attr == 'L':
                    x, y = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.pendown()
                    turtle.goto(-self.width/2+x, self.height/2-y)
                    turtle.penup()
                elif attr == 'l':
                    dx, dy = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.pendown()
                    turtle.goto(turtle.xcor()+dx, turtle.ycor()-dy)
                    turtle.penup()
                    endl = attr
                elif endl == 'C':
                    p1 = attr * scale[0], attrs_yield.__next__() * scale[1]
                    p2 = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    p3 = attrs_yield.__next__() * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.penup()
                    p0 = turtle.xcor() + self.width / 2, self.height / 2 - turtle.ycor()
                    self.ThirdOrderBezier(p0, p1, p2, p3)
                elif endl == 'c':
                    turtle.penup()
                    p0 = turtle.xcor() + self.width / 2, self.height / 2 - turtle.ycor()
                    p1 = attr * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    p2 = attrs_yield.__next__() * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    p3 = attrs_yield.__next__() * scale[0] + p0[0], attrs_yield.__next__() * scale[1] + p0[1]
                    self.ThirdOrderBezier(p0, p1, p2, p3)
                elif endl == 'L':
                    x, y = attr * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.pendown()
                    turtle.goto(-self.width/2+x, self.height/2-y)
                    turtle.penup()
                elif endl == 'l':
                    dx, dy = attr * scale[0], attrs_yield.__next__() * scale[1]
                    turtle.pendown()
                    turtle.goto(turtle.xcor()+dx, turtle.ycor()-dy)
                    turtle.penup()
        turtle.penup()
        turtle.hideturtle()
        turtle.update()
        svgfile.close()
    '''attrs生成器'''
    @staticmethod
    def yieldattrs(attrs):
        for attr in attrs:
            if attr.isdigit():
                yield float(attr)
            elif attr[0].isalpha():
                yield attr[0]
                yield float(attr[1:])
            elif attr[-1].isalpha():
                yield float(attr[0: -1])
            elif attr[0] == '-':
                yield float(attr)


'''
Function:
    利用遗传算法画画
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import cv2
import copy
import random
import numpy as np
from ...utils import checkdir
from PIL import Image, ImageDraw
from ..base import BaseBeautifier


'''圆形'''
class Circle():
    def __init__(self, radius_range=50, radius_shift_range=50, center_shift_range=50, color_shift_range=50, target_image=None):
        # 属性
        self.radius_range = radius_range
        self.radius_shift_range = radius_shift_range
        self.center_shift_range = center_shift_range
        self.color_shift_range = color_shift_range
        self.target_image = target_image
        # 中心点
        image_width, image_height = target_image.size[:2]
        x, y = random.randint(0, int(image_width)), random.randint(0, int(image_height))
        self.image_width, self.image_height = image_width, image_height
        self.center = (x, y)
        # 半径
        self.radius = random.randint(1, radius_range)
        # 颜色
        r, g, b = np.asarray(target_image)[min(y, image_height-1), min(x, image_width-1)]
        self.color = (int(r), int(g), int(b), random.randint(0, 256))
    '''变异'''
    def mutate(self):
        mutations = ['center', 'radius', 'color', 'reset']
        mutation_type = random.choice(mutations)
        # 随机中心的偏移
        if mutation_type == 'center':
            x_shift = int(random.randint(-self.center_shift_range, self.center_shift_range) * random.random())
            y_shift = int(random.randint(-self.center_shift_range, self.center_shift_range) * random.random())
            x = min(max(0, x_shift + self.center[0]), self.image_width)
            y = min(max(0, y_shift + self.center[1]), self.image_height)
            self.center = (x, y)
        # 随机改变半径
        elif mutation_type == 'radius':
            self.radius += int(random.randint(-self.radius_shift_range, self.radius_shift_range) * random.random())
        # 随机改变颜色
        elif mutation_type == 'color':
            self.color = tuple(c + int(random.randint(-self.color_shift_range, self.color_shift_range) * random.random()) for c in self.color)
            self.color = tuple(min(max(c, 0), 255) for c in self.color)
        # 重置
        else:
            new_circle = Circle(
                radius_range=self.radius_range, 
                radius_shift_range=self.radius_shift_range, 
                center_shift_range=self.center_shift_range, 
                color_shift_range=self.color_shift_range, 
                target_image=self.target_image,
            )
            self.center = new_circle.center
            self.radius = new_circle.radius
            self.color = new_circle.color


'''利用遗传算法画画'''
class GeneticFittingCircleBeautifier(BaseBeautifier):
    def __init__(self, init_cfg=None, cache_dir='cache', save_cache=True, **kwargs):
        super(GeneticFittingCircleBeautifier, self).__init__(**kwargs)
        if init_cfg is None:
            init_cfg = {
                'num_populations': 10,
                'init_num_circles': 1,
                'num_generations': 1e5,
                'print_interval': 1,
                'mutation_rate': 0.1,
                'selection_rate': 0.5,
                'crossover_rate': 0.5,
                'circle_cfg': {'radius_range': 50, 'radius_shift_range': 50, 'center_shift_range': 50, 'color_shift_range': 50},
            }
        self.init_cfg = init_cfg
        self.cache_dir = cache_dir
        self.save_cache = save_cache
    '''迭代图片'''
    def iterimage(self, image):
        image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        # 初始化
        populations = []
        for _ in range(self.init_cfg['num_populations']):
            population = []
            for _ in range(self.init_cfg['init_num_circles']):
                population.append(Circle(
                    target_image=image, 
                    **self.init_cfg['circle_cfg']
                ))
            populations.append(population)
        # 迭代
        mutation_rate = self.init_cfg['mutation_rate']
        for g in range(1, int(self.init_cfg['num_generations']+1)):
            fitnesses = []
            for idx, population in enumerate(copy.deepcopy(populations)):
                fitness_ori = self.calcfitnesses([population], image)[0]
                fitness = 0
                while fitness_ori > fitness:
                    population_new = population + [Circle(target_image=image, **self.init_cfg['circle_cfg'])]
                    fitness = self.calcfitnesses([population_new], image)[0]
                populations[idx] = population_new
                fitnesses.append(fitness)
            if g % self.init_cfg['print_interval'] == 0:
                if self.save_cache:
                    population = populations[np.argmax(fitnesses)]
                    output_image = self.draw(population, image)
                    checkdir(self.cache_dir)
                    output_image.save(os.path.join(self.cache_dir, f'cache_g{g}.png'))
                self.logger_handle.info(f'Generation: {g}, FITNESS: {max(fitnesses)}')
            num_populations = len(populations)
            # --自然选择
            populations = self.select(image, fitnesses, populations)
            # --交叉
            populations = self.crossover(populations, num_populations)
            # --变异
            populations = self.mutate(image, populations, mutation_rate)
        # 选择最优解返回
        population = populations[np.argmax(fitnesses)]
        output_image = self.draw(population, image)
        return cv2.cvtColor(np.asarray(output_image), cv2.COLOR_RGB2BGR)
    '''自然选择'''
    def select(self, image, fitnesses, populations):
        sorted_idx = np.argsort(fitnesses)[::-1]
        selected_populations = []
        selection_rate = self.init_cfg['selection_rate']
        for idx in range(int(len(populations) * selection_rate)):
            selected_idx = int(sorted_idx[idx])
            selected_populations.append(populations[selected_idx])
        return selected_populations
    '''交叉'''
    def crossover(self, populations, num_populations):
        indices = list(range(len(populations)))
        while len(populations) < num_populations:
            idx1 = random.choice(indices)
            idx2 = random.choice(indices)
            population1 = copy.deepcopy(populations[idx1])
            population2 = copy.deepcopy(populations[idx2])
            for circle_idx in range(len(population1)):
                if self.init_cfg['crossover_rate'] > random.random():
                    population1[circle_idx] = population2[circle_idx]
            populations.append(population1)
        return populations
    '''变异'''
    def mutate(self, target_image, populations, mutation_rate):
        populations_new = copy.deepcopy(populations)
        for idx, population in enumerate(populations):
            fitness_ori = self.calcfitnesses([population], target_image)[0]
            for circle in population:
                if mutation_rate > random.random():
                    circle.mutate()
            fitness = self.calcfitnesses([population], target_image)[0]
            if fitness > fitness_ori: populations_new[idx] = population
        return populations_new
    '''计算适应度'''
    def calcfitnesses(self, populations, target_image):
        fitnesses = []
        for idx in range(len(populations)):
            image = self.draw(populations[idx], target_image)
            fitnesses.append(self.calcsimilarity(image, target_image))
        return fitnesses
    '''画图'''
    def draw(self, population, target_image):
        image = Image.new('RGB', target_image.size, "#FFFFFF")
        for circle in population:
            item = Image.new('RGB', target_image.size, "#000000")
            mask = Image.new('L', target_image.size, 255)
            draw = ImageDraw.Draw(item)
            draw.ellipse((circle.center[0]-circle.radius, circle.center[1]-circle.radius, circle.center[0]+circle.radius, circle.center[1]+circle.radius), fill=circle.color)
            draw = ImageDraw.Draw(mask)
            draw.ellipse((circle.center[0]-circle.radius, circle.center[1]-circle.radius, circle.center[0]+circle.radius, circle.center[1]+circle.radius), fill=128)
            image = Image.composite(image, item, mask)
        return image
    '''计算两幅图之间的相似度'''
    def calcsimilarity(self, image, target_image):
        image = np.asarray(image) / 255.
        target_image = np.asarray(target_image) / 255.
        similarity = 1 - abs(image - target_image).mean()
        return similarity


'''
Function:
    一些终端打印工具
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import logging


'''打印日志类'''
class Logger():
    def __init__(self, logfilepath, **kwargs):
        setattr(self, 'logfilepath', logfilepath)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)-8s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=[logging.FileHandler(logfilepath), logging.StreamHandler()],
        )
    @staticmethod
    def log(level, message):
        logging.log(level, message)
    def debug(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.DEBUG, message)
    def info(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.INFO, message)
    def warning(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.WARNING, message)
    def error(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.ERROR, message)


'''
Function:
    复现论文"Perceptual Losses for Real-Time Style Transfer and Super-Resolution"
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
import torch.utils.model_zoo as model_zoo
import torchvision.transforms as transforms
from ..base import BaseBeautifier


'''ConvBlock'''
class ConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, upsample=False, normalize=True, relu=True):
        super(ConvBlock, self).__init__()
        self.upsample = upsample
        self.block = nn.Sequential(
            nn.ReflectionPad2d(kernel_size // 2), nn.Conv2d(in_channels, out_channels, kernel_size, stride)
        )
        self.norm = nn.InstanceNorm2d(out_channels, affine=True) if normalize else None
        self.relu = relu
    '''forward'''
    def forward(self, x):
        if self.upsample: x = F.interpolate(x, scale_factor=2)
        x = self.block(x)
        if self.norm is not None: x = self.norm(x)
        if self.relu: x = F.relu(x)
        return x


'''ResidualBlock'''
class ResidualBlock(nn.Module):
    def __init__(self, channels):
        super(ResidualBlock, self).__init__()
        self.block = nn.Sequential(
            ConvBlock(channels, channels, kernel_size=3, stride=1, normalize=True, relu=True),
            ConvBlock(channels, channels, kernel_size=3, stride=1, normalize=True, relu=False),
        )
    '''forward'''
    def forward(self, x):
        return self.block(x) + x


'''TransformerNet, 模型修改自: https://github.com/eriklindernoren/Fast-Neural-Style-Transfer'''
class TransformerNet(nn.Module):
    def __init__(self):
        super(TransformerNet, self).__init__()
        self.model = nn.Sequential(
            ConvBlock(3, 32, kernel_size=9, stride=1),
            ConvBlock(32, 64, kernel_size=3, stride=2),
            ConvBlock(64, 128, kernel_size=3, stride=2),
            ResidualBlock(128),
            ResidualBlock(128),
            ResidualBlock(128),
            ResidualBlock(128),
            ResidualBlock(128),
            ConvBlock(128, 64, kernel_size=3, upsample=True),
            ConvBlock(64, 32, kernel_size=3, upsample=True),
            ConvBlock(32, 3, kernel_size=9, stride=1, normalize=False, relu=False),
        )
    '''forward'''
    def forward(self, x):
        return self.model(x)


'''复现论文"Perceptual Losses for Real-Time Style Transfer and Super-Resolution"'''
class FastNeuralStyleTransferBeautifier(BaseBeautifier):
    def __init__(self, style='starrynight', use_cuda=True, **kwargs):
        super(FastNeuralStyleTransferBeautifier, self).__init__(**kwargs)
        self.model_urls = {
            'cuphead': 'https://github.com/CharlesPikachu/pydrawing/releases/download/checkpoints/fastneuralstyletransfer_cuphead.pth',
            'mosaic': 'https://github.com/CharlesPikachu/pydrawing/releases/download/checkpoints/fastneuralstyletransfer_mosaic.pth',
            'starrynight': 'https://github.com/CharlesPikachu/pydrawing/releases/download/checkpoints/fastneuralstyletransfer_starrynight.pth',
        }
        self.mean = np.array([0.485, 0.456, 0.406])
        self.std = np.array([0.229, 0.224, 0.225])
        self.preprocess = transforms.Compose([transforms.ToTensor(), transforms.Normalize(self.mean, self.std)])
        assert style in self.model_urls
        self.style = style
        self.use_cuda = use_cuda
        self.transformer = TransformerNet()
        self.transformer.load_state_dict(model_zoo.load_url(self.model_urls[style]))
        self.transformer.eval()
        if torch.cuda.is_available() and self.use_cuda: self.transformer = self.transformer.cuda()
    '''迭代图片'''
    def iterimage(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        input_image = self.preprocess(image).unsqueeze(0)
        if torch.cuda.is_available() and self.use_cuda: 
            input_image = input_image.cuda()
        with torch.no_grad(): 
            output_image = self.transformer(input_image)[0]
        output_image = output_image.data.cpu().float()
        for c in range(3):
            output_image[c, :].mul_(self.std[c]).add_(self.mean[c])
        output_image = output_image.mul(255).add_(0.5).clamp_(0, 255).permute(1, 2, 0).to('cpu', torch.uint8).numpy()
        output_image = cv2.cvtColor(output_image, cv2.COLOR_RGB2BGR)
        output_image = cv2.resize(output_image, (image.shape[1], image.shape[0]))
        return output_image


'''
Function:
    信号故障的效果
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import random
from ...utils import checkdir
from ..base import BaseBeautifier


'''信号故障的效果'''
class GlitchBeautifier(BaseBeautifier):
    def __init__(self, header_size=200, intensity=0.1, block_size=100, **kwargs):
        super(GlitchBeautifier, self).__init__(**kwargs)
        self.header_size, self.intensity, self.block_size = header_size, intensity, block_size
    '''迭代图片'''
    def process(self, filepath):
        checkdir(self.savedir)
        ext = filepath.split('.')[-1]
        assert ext.lower() in ['mp4', 'avi']
        with open(filepath, 'rb') as fp_in:
            with open(os.path.join(self.savedir, f'{self.savename}.{ext}'), 'wb') as fp_out:
                fp_out.write(fp_in.read(self.header_size))
                while True:
                    block_data = fp_in.read(self.block_size)
                    if not block_data: break
                    if random.random() < self.intensity / 100: block_data = os.urandom(self.block_size)
                    fp_out.write(block_data)
        self.logger_handle.info(f'Video is saved into {self.savename}.{ext}')


'''
Function:
    用Python美化你的照片或视频
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import warnings
if __name__ == '__main__':
    from modules import *
else:
    from .modules import *
warnings.filterwarnings('ignore')


'''用Python美化你的照片或视频'''
class pydrawing():
    def __init__(self, **kwargs):
        for key, value in kwargs.items(): setattr(self, key, value)
        self.supported_beautifiers = self.initializebeautifiers()
        self.logger_handle = Logger(kwargs.get('logfilepath', 'pydrawing.log'))
    '''执行对应的算法'''
    def execute(self, filepath='asserts/dog.jpg', beautifier_type=None, config={}):
        assert beautifier_type in self.supported_beautifiers, 'unsupport beautifier_type %s' % beautifier_type
        if 'savedir' not in config: config['savedir'] = 'outputs'
        if 'savename' not in config: config['savename'] = 'output'
        if 'logger_handle' not in config: config['logger_handle'] = self.logger_handle
        beautifier = self.supported_beautifiers[beautifier_type](**config)
        beautifier.process(filepath)
    '''获得所有支持的美化器'''
    def getallsupports(self):
        return list(self.supported_beautifiers.keys())
    '''初始化美化器'''
    def initializebeautifiers(self):
        supported_beautifiers = {
            'glitch': GlitchBeautifier,
            'cartoonise': CartooniseBeautifier,
            'cartoongan': CartoonGanBeautifier,
            'beziercurve': BezierCurveBeautifier,
            'photomosaic': PhotomosaicBeautifier,
            'characterize': CharacterizeBeautifier,
            'douyineffect': DouyinEffectBeautifier,
            'pencildrawing': PencilDrawingBeautifier,
            'cartoonizeface': CartoonizeFaceBeautifier,
            'geneticfittingcircle': GeneticFittingCircleBeautifier,
            'geneticfittingpolygon': GeneticFittingPolygonBeautifier,
            'fastneuralstyletransfer': FastNeuralStyleTransferBeautifier,
        }
        return supported_beautifiers
    '''repr'''
    def __repr__(self):
        return 'pydrawing, author: charles, 微信公众号: Charles的皮卡丘'


'''run'''
if __name__ == '__main__':
    import random
    drawing_client = pydrawing()
    drawing_client.execute('asserts/dog.jpg', random.choice(drawing_client.getallsupports()))


'''
Function:
    拼马赛克图片
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import glob
import numpy as np
from tqdm import tqdm
from itertools import product
from ..base import BaseBeautifier


'''拼马赛克图片'''
class PhotomosaicBeautifier(BaseBeautifier):
    def __init__(self, block_size=15, src_images_dir=None, **kwargs):
        super(PhotomosaicBeautifier, self).__init__(**kwargs)
        self.block_size = block_size
        self.src_images_dir = src_images_dir
        self.src_images, self.avg_colors = self.ReadSourceImages()
    '''迭代图片'''
    def iterimage(self, image):
        output_image = np.zeros(image.shape, np.uint8)
        src_images, avg_colors = self.src_images, self.avg_colors
        for i, j in tqdm(product(range(int(image.shape[1]/self.block_size)), range(int(image.shape[0]/self.block_size)))):
            block = image[j*self.block_size: (j+1)*self.block_size, i*self.block_size: (i+1)*self.block_size, :]
            avg_color = np.sum(np.sum(block, axis=0), axis=0) / (self.block_size * self.block_size)
            distances = np.linalg.norm(avg_color - avg_colors, axis=1)
            idx = np.argmin(distances)
            output_image[j*self.block_size: (j+1)*self.block_size, i*self.block_size: (i+1)*self.block_size, :] = src_images[idx]
        return output_image
    '''读取所有源图片并计算对应的颜色平均值'''
    def ReadSourceImages(self):
        src_images, avg_colors = [], []
        for path in tqdm(glob.glob("{}/*.jpg".format(self.src_images_dir))):
            image = cv2.imread(path, cv2.IMREAD_COLOR)
            if image.shape[-1] != 3: continue
            image = cv2.resize(image, (self.block_size, self.block_size))
            avg_color = np.sum(np.sum(image, axis=0), axis=0) / (self.block_size * self.block_size)
            src_images.append(image)
            avg_colors.append(avg_color)
        return src_images, np.array(avg_colors)


'''
Function:
    复现论文"CartoonGAN: Generative Adversarial Networks for Photo Cartoonization"
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
import torchvision.transforms as transforms
from ..base import BaseBeautifier


'''instance normalization'''
class InstanceNormalization(nn.Module):
    def __init__(self, dim, eps=1e-9):
        super(InstanceNormalization, self).__init__()
        self.scale = nn.Parameter(torch.FloatTensor(dim))
        self.shift = nn.Parameter(torch.FloatTensor(dim))
        self.eps = eps
    '''call'''
    def __call__(self, x):
        n = x.size(2) * x.size(3)
        t = x.view(x.size(0), x.size(1), n)
        mean = torch.mean(t, 2).unsqueeze(2).unsqueeze(3).expand_as(x)
        var = torch.var(t, 2).unsqueeze(2).unsqueeze(3).expand_as(x) * ((n - 1) / float(n))
        scale_broadcast = self.scale.unsqueeze(1).unsqueeze(1).unsqueeze(0)
        scale_broadcast = scale_broadcast.expand_as(x)
        shift_broadcast = self.shift.unsqueeze(1).unsqueeze(1).unsqueeze(0)
        shift_broadcast = shift_broadcast.expand_as(x)
        out = (x - mean) / torch.sqrt(var + self.eps)
        out = out * scale_broadcast + shift_broadcast
        return out


'''网络结构, 模型修改自: https://github.com/Yijunmaverick/CartoonGAN-Test-Pytorch-Torch'''
class Transformer(nn.Module):
    def __init__(self):
        super(Transformer, self).__init__()
        self.refpad01_1 = nn.ReflectionPad2d(3)
        self.conv01_1 = nn.Conv2d(3, 64, 7)
        self.in01_1 = InstanceNormalization(64)
        self.conv02_1 = nn.Conv2d(64, 128, 3, 2, 1)
        self.conv02_2 = nn.Conv2d(128, 128, 3, 1, 1)
        self.in02_1 = InstanceNormalization(128)
        self.conv03_1 = nn.Conv2d(128, 256, 3, 2, 1)
        self.conv03_2 = nn.Conv2d(256, 256, 3, 1, 1)
        self.in03_1 = InstanceNormalization(256)    
        self.refpad04_1 = nn.ReflectionPad2d(1)
        self.conv04_1 = nn.Conv2d(256, 256, 3)
        self.in04_1 = InstanceNormalization(256)
        self.refpad04_2 = nn.ReflectionPad2d(1)
        self.conv04_2 = nn.Conv2d(256, 256, 3)
        self.in04_2 = InstanceNormalization(256)
        self.refpad05_1 = nn.ReflectionPad2d(1)
        self.conv05_1 = nn.Conv2d(256, 256, 3)
        self.in05_1 = InstanceNormalization(256)
        self.refpad05_2 = nn.ReflectionPad2d(1)
        self.conv05_2 = nn.Conv2d(256, 256, 3)
        self.in05_2 = InstanceNormalization(256)
        self.refpad06_1 = nn.ReflectionPad2d(1)
        self.conv06_1 = nn.Conv2d(256, 256, 3)
        self.in06_1 = InstanceNormalization(256)
        self.refpad06_2 = nn.ReflectionPad2d(1)
        self.conv06_2 = nn.Conv2d(256, 256, 3)
        self.in06_2 = InstanceNormalization(256)
        self.refpad07_1 = nn.ReflectionPad2d(1)
        self.conv07_1 = nn.Conv2d(256, 256, 3)
        self.in07_1 = InstanceNormalization(256)
        self.refpad07_2 = nn.ReflectionPad2d(1)
        self.conv07_2 = nn.Conv2d(256, 256, 3)
        self.in07_2 = InstanceNormalization(256)
        self.refpad08_1 = nn.ReflectionPad2d(1)
        self.conv08_1 = nn.Conv2d(256, 256, 3)
        self.in08_1 = InstanceNormalization(256)
        self.refpad08_2 = nn.ReflectionPad2d(1)
        self.conv08_2 = nn.Conv2d(256, 256, 3)
        self.in08_2 = InstanceNormalization(256)
        self.refpad09_1 = nn.ReflectionPad2d(1)
        self.conv09_1 = nn.Conv2d(256, 256, 3)
        self.in09_1 = InstanceNormalization(256)
        self.refpad09_2 = nn.ReflectionPad2d(1)
        self.conv09_2 = nn.Conv2d(256, 256, 3)
        self.in09_2 = InstanceNormalization(256)
        self.refpad10_1 = nn.ReflectionPad2d(1)
        self.conv10_1 = nn.Conv2d(256, 256, 3)
        self.in10_1 = InstanceNormalization(256)
        self.refpad10_2 = nn.ReflectionPad2d(1)
        self.conv10_2 = nn.Conv2d(256, 256, 3)
        self.in10_2 = InstanceNormalization(256)
        self.refpad11_1 = nn.ReflectionPad2d(1)
        self.conv11_1 = nn.Conv2d(256, 256, 3)
        self.in11_1 = InstanceNormalization(256)
        self.refpad11_2 = nn.ReflectionPad2d(1)
        self.conv11_2 = nn.Conv2d(256, 256, 3)
        self.in11_2 = InstanceNormalization(256)
        self.deconv01_1 = nn.ConvTranspose2d(256, 128, 3, 2, 1, 1)
        self.deconv01_2 = nn.Conv2d(128, 128, 3, 1, 1)
        self.in12_1 = InstanceNormalization(128)
        self.deconv02_1 = nn.ConvTranspose2d(128, 64, 3, 2, 1, 1)
        self.deconv02_2 = nn.Conv2d(64, 64, 3, 1, 1)
        self.in13_1 = InstanceNormalization(64)
        self.refpad12_1 = nn.ReflectionPad2d(3)
        self.deconv03_1 = nn.Conv2d(64, 3, 7)
    '''forward'''
    def forward(self, x):
        y = F.relu(self.in01_1(self.conv01_1(self.refpad01_1(x))))
        y = F.relu(self.in02_1(self.conv02_2(self.conv02_1(y))))
        t04 = F.relu(self.in03_1(self.conv03_2(self.conv03_1(y))))
        y = F.relu(self.in04_1(self.conv04_1(self.refpad04_1(t04))))
        t05 = self.in04_2(self.conv04_2(self.refpad04_2(y))) + t04
        y = F.relu(self.in05_1(self.conv05_1(self.refpad05_1(t05))))
        t06 = self.in05_2(self.conv05_2(self.refpad05_2(y))) + t05
        y = F.relu(self.in06_1(self.conv06_1(self.refpad06_1(t06))))
        t07 = self.in06_2(self.conv06_2(self.refpad06_2(y))) + t06
        y = F.relu(self.in07_1(self.conv07_1(self.refpad07_1(t07))))
        t08 = self.in07_2(self.conv07_2(self.refpad07_2(y))) + t07
        y = F.relu(self.in08_1(self.conv08_1(self.refpad08_1(t08))))
        t09 = self.in08_2(self.conv08_2(self.refpad08_2(y))) + t08
        y = F.relu(self.in09_1(self.conv09_1(self.refpad09_1(t09))))
        t10 = self.in09_2(self.conv09_2(self.refpad09_2(y))) + t09
        y = F.relu(self.in10_1(self.conv10_1(self.refpad10_1(t10))))
        t11 = self.in10_2(self.conv10_2(self.refpad10_2(y))) + t10
        y = F.relu(self.in11_1(self.conv11_1(self.refpad11_1(t11))))
        y = self.in11_2(self.conv11_2(self.refpad11_2(y))) + t11
        y = F.relu(self.in12_1(self.deconv01_2(self.deconv01_1(y))))
        y = F.relu(self.in13_1(self.deconv02_2(self.deconv02_1(y))))
        y = F.tanh(self.deconv03_1(self.refpad12_1(y)))
        return y


'''复现论文"CartoonGAN: Generative Adversarial Networks for Photo Cartoonization"'''
class CartoonGanBeautifier(BaseBeautifier):
    def __init__(self, style='Hosoda', use_cuda=True, **kwargs):
        super(CartoonGanBeautifier, self).__init__(**kwargs)
        self.model_urls = {
            'Hayao': 'http://vllab1.ucmerced.edu/~yli62/CartoonGAN/pytorch_pth/Hayao_net_G_float.pth',
            'Hosoda': 'http://vllab1.ucmerced.edu/~yli62/CartoonGAN/pytorch_pth/Hosoda_net_G_float.pth',
            'Paprika': 'http://vllab1.ucmerced.edu/~yli62/CartoonGAN/pytorch_pth/Paprika_net_G_float.pth',
            'Shinkai': 'http://vllab1.ucmerced.edu/~yli62/CartoonGAN/pytorch_pth/Shinkai_net_G_float.pth',
        }
        assert style in self.model_urls
        self.style = style
        self.use_cuda = use_cuda
        self.transformer = Transformer()
        self.transformer.load_state_dict(model_zoo.load_url(self.model_urls[style]))
        self.transformer.eval()
        if torch.cuda.is_available() and self.use_cuda: self.transformer = self.transformer.cuda()
    '''迭代图片'''
    def iterimage(self, image):
        input_image = transforms.ToTensor()(image).unsqueeze(0)
        input_image = -1 + 2 * input_image
        if torch.cuda.is_available() and self.use_cuda: 
            input_image = input_image.cuda()
        with torch.no_grad(): 
            output_image = self.transformer(input_image)[0]
        output_image = output_image.data.cpu().float() * 0.5 + 0.5
        output_image = (output_image.numpy() * 255).astype(np.uint8)
        output_image = np.transpose(output_image, (1, 2, 0))
        output_image = cv2.resize(output_image, (image.shape[1], image.shape[0]))
        return output_image


'''
Function:
    setup the pydrawing
Author:
    Charles
微信公众号:
    Charles的皮卡丘
GitHub:
    https://github.com/CharlesPikachu
'''
import pydrawing
from setuptools import setup, find_packages


'''readme'''
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


'''package data'''
package_data = {}
package_data.update({
    'pydrawing.modules.beautifiers.pencildrawing': ['textures/*'] 
})
package_data.update({
    'pydrawing.modules.beautifiers.beziercurve': ['potrace.exe'] 
})


'''setup'''
setup(
    name=pydrawing.__title__,
    version=pydrawing.__version__,
    description=pydrawing.__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ],
    author=pydrawing.__author__,
    url=pydrawing.__url__,
    author_email=pydrawing.__email__,
    license=pydrawing.__license__,
    include_package_data=True,
    package_data=package_data,
    install_requires=list(open('requirements.txt', 'r').readlines()),
    zip_safe=True,
    packages=find_packages(),
)


'''
Function:
    人脸卡通化
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo
from PIL import Image
from ..base import BaseBeautifier
from .facedetector import FaceDetector
from .facesegmentor import FaceSegmentor
from torch.nn.parameter import Parameter


'''ConvBlock'''
class ConvBlock(nn.Module):
    def __init__(self, dim_in, dim_out):
        super(ConvBlock, self).__init__()
        self.dim_out = dim_out
        self.ConvBlock1 = nn.Sequential(
            nn.InstanceNorm2d(dim_in),
            nn.ReLU(True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(dim_in, dim_out//2, kernel_size=3, stride=1, bias=False)
        )
        self.ConvBlock2 = nn.Sequential(
            nn.InstanceNorm2d(dim_out//2),
            nn.ReLU(True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(dim_out//2, dim_out//4, kernel_size=3, stride=1, bias=False)
        )
        self.ConvBlock3 = nn.Sequential(
            nn.InstanceNorm2d(dim_out//4),
            nn.ReLU(True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(dim_out//4, dim_out//4, kernel_size=3, stride=1, bias=False)
        )
        self.ConvBlock4 = nn.Sequential(
            nn.InstanceNorm2d(dim_in),
            nn.ReLU(True),
            nn.Conv2d(dim_in, dim_out, kernel_size=1, stride=1, bias=False)
        )
    '''forward'''
    def forward(self, x):
        residual = x
        x1 = self.ConvBlock1(x)
        x2 = self.ConvBlock2(x1)
        x3 = self.ConvBlock3(x2)
        out = torch.cat((x1, x2, x3), 1)
        if residual.size(1) != self.dim_out: residual = self.ConvBlock4(residual)
        return residual + out


'''HourGlassBlock'''
class HourGlassBlock(nn.Module):
    def __init__(self, dim_in, dim_out):
        super(HourGlassBlock, self).__init__()
        self.ConvBlock1_1 = ConvBlock(dim_in, dim_out)
        self.ConvBlock1_2 = ConvBlock(dim_out, dim_out)
        self.ConvBlock2_1 = ConvBlock(dim_out, dim_out)
        self.ConvBlock2_2 = ConvBlock(dim_out, dim_out)
        self.ConvBlock3_1 = ConvBlock(dim_out, dim_out)
        self.ConvBlock3_2 = ConvBlock(dim_out, dim_out)
        self.ConvBlock4_1 = ConvBlock(dim_out, dim_out)
        self.ConvBlock4_2 = ConvBlock(dim_out, dim_out)
        self.ConvBlock5 = ConvBlock(dim_out, dim_out)
        self.ConvBlock6 = ConvBlock(dim_out, dim_out)
        self.ConvBlock7 = ConvBlock(dim_out, dim_out)
        self.ConvBlock8 = ConvBlock(dim_out, dim_out)
        self.ConvBlock9 = ConvBlock(dim_out, dim_out)
    '''forward'''
    def forward(self, x):
        skip1 = self.ConvBlock1_1(x)
        down1 = F.avg_pool2d(x, 2)
        down1 = self.ConvBlock1_2(down1)
        skip2 = self.ConvBlock2_1(down1)
        down2 = F.avg_pool2d(down1, 2)
        down2 = self.ConvBlock2_2(down2)
        skip3 = self.ConvBlock3_1(down2)
        down3 = F.avg_pool2d(down2, 2)
        down3 = self.ConvBlock3_2(down3)
        skip4 = self.ConvBlock4_1(down3)
        down4 = F.avg_pool2d(down3, 2)
        down4 = self.ConvBlock4_2(down4)
        center = self.ConvBlock5(down4)
        up4 = self.ConvBlock6(center)
        up4 = F.upsample(up4, scale_factor=2)
        up4 = skip4 + up4
        up3 = self.ConvBlock7(up4)
        up3 = F.upsample(up3, scale_factor=2)
        up3 = skip3 + up3
        up2 = self.ConvBlock8(up3)
        up2 = F.upsample(up2, scale_factor=2)
        up2 = skip2 + up2
        up1 = self.ConvBlock9(up2)
        up1 = F.upsample(up1, scale_factor=2)
        up1 = skip1 + up1
        return up1


'''HourGlass'''
class HourGlass(nn.Module):
    def __init__(self, dim_in, dim_out, use_res=True):
        super(HourGlass, self).__init__()
        self.use_res = use_res
        self.HG = nn.Sequential(
            HourGlassBlock(dim_in, dim_out),
            ConvBlock(dim_out, dim_out),
            nn.Conv2d(dim_out, dim_out, kernel_size=1, stride=1, bias=False),
            nn.InstanceNorm2d(dim_out),
            nn.ReLU(True)
        )
        self.Conv1 = nn.Conv2d(dim_out, 3, kernel_size=1, stride=1)
        if self.use_res:
            self.Conv2 = nn.Conv2d(dim_out, dim_out, kernel_size=1, stride=1)
            self.Conv3 = nn.Conv2d(3, dim_out, kernel_size=1, stride=1)
    '''forward'''
    def forward(self, x):
        ll = self.HG(x)
        tmp_out = self.Conv1(ll)
        if self.use_res:
            ll = self.Conv2(ll)
            tmp_out_ = self.Conv3(tmp_out)
            return x + ll + tmp_out_
        else:
            return tmp_out


'''ResnetBlock'''
class ResnetBlock(nn.Module):
    def __init__(self, dim, use_bias=False):
        super(ResnetBlock, self).__init__()
        conv_block = []
        conv_block += [nn.ReflectionPad2d(1), nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=0, bias=use_bias), nn.InstanceNorm2d(dim), nn.ReLU(True)]
        conv_block += [nn.ReflectionPad2d(1), nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=0, bias=use_bias), nn.InstanceNorm2d(dim)]
        self.conv_block = nn.Sequential(*conv_block)
    '''forward'''
    def forward(self, x):
        out = x + self.conv_block(x)
        return out


'''adaLIN'''
class adaLIN(nn.Module):
    def __init__(self, num_features, eps=1e-5):
        super(adaLIN, self).__init__()
        self.eps = eps
        self.rho = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.rho.data.fill_(0.9)
    '''forward'''
    def forward(self, input, gamma, beta):
        in_mean, in_var = torch.mean(input, dim=[2, 3], keepdim=True), torch.var(input, dim=[2, 3], keepdim=True)
        out_in = (input - in_mean) / torch.sqrt(in_var + self.eps)
        ln_mean, ln_var = torch.mean(input, dim=[1, 2, 3], keepdim=True), torch.var(input, dim=[1, 2, 3], keepdim=True)
        out_ln = (input - ln_mean) / torch.sqrt(ln_var + self.eps)
        out = self.rho.expand(input.shape[0], -1, -1, -1) * out_in + (1-self.rho.expand(input.shape[0], -1, -1, -1)) * out_ln
        out = out * gamma.unsqueeze(2).unsqueeze(3) + beta.unsqueeze(2).unsqueeze(3)
        return out


'''SoftAdaLIN'''
class SoftAdaLIN(nn.Module):
    def __init__(self, num_features, eps=1e-5):
        super(SoftAdaLIN, self).__init__()
        self.norm = adaLIN(num_features, eps)
        self.w_gamma = Parameter(torch.zeros(1, num_features))
        self.w_beta = Parameter(torch.zeros(1, num_features))
        self.c_gamma = nn.Sequential(nn.Linear(num_features, num_features), nn.ReLU(True), nn.Linear(num_features, num_features))
        self.c_beta = nn.Sequential(nn.Linear(num_features, num_features), nn.ReLU(True), nn.Linear(num_features, num_features))
        self.s_gamma = nn.Linear(num_features, num_features)
        self.s_beta = nn.Linear(num_features, num_features)
    '''forward'''
    def forward(self, x, content_features, style_features):
        content_gamma, content_beta = self.c_gamma(content_features), self.c_beta(content_features)
        style_gamma, style_beta = self.s_gamma(style_features), self.s_beta(style_features)
        w_gamma, w_beta = self.w_gamma.expand(x.shape[0], -1), self.w_beta.expand(x.shape[0], -1)
        soft_gamma = (1. - w_gamma) * style_gamma + w_gamma * content_gamma
        soft_beta = (1. - w_beta) * style_beta + w_beta * content_beta
        out = self.norm(x, soft_gamma, soft_beta)
        return out


'''ResnetSoftAdaLINBlock'''
class ResnetSoftAdaLINBlock(nn.Module):
    def __init__(self, dim, use_bias=False):
        super(ResnetSoftAdaLINBlock, self).__init__()
        self.pad1 = nn.ReflectionPad2d(1)
        self.conv1 = nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=0, bias=use_bias)
        self.norm1 = SoftAdaLIN(dim)
        self.relu1 = nn.ReLU(True)
        self.pad2 = nn.ReflectionPad2d(1)
        self.conv2 = nn.Conv2d(dim, dim, kernel_size=3, stride=1, padding=0, bias=use_bias)
        self.norm2 = SoftAdaLIN(dim)
    '''forward'''
    def forward(self, x, content_features, style_features):
        out = self.pad1(x)
        out = self.conv1(out)
        out = self.norm1(out, content_features, style_features)
        out = self.relu1(out)
        out = self.pad2(out)
        out = self.conv2(out)
        out = self.norm2(out, content_features, style_features)
        return out + x


'''LIN'''
class LIN(nn.Module):
    def __init__(self, num_features, eps=1e-5):
        super(LIN, self).__init__()
        self.eps = eps
        self.rho = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.gamma = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.beta = Parameter(torch.Tensor(1, num_features, 1, 1))
        self.rho.data.fill_(0.0)
        self.gamma.data.fill_(1.0)
        self.beta.data.fill_(0.0)
    '''forward'''
    def forward(self, input):
        in_mean, in_var = torch.mean(input, dim=[2, 3], keepdim=True), torch.var(input, dim=[2, 3], keepdim=True)
        out_in = (input - in_mean) / torch.sqrt(in_var + self.eps)
        ln_mean, ln_var = torch.mean(input, dim=[1, 2, 3], keepdim=True), torch.var(input, dim=[1, 2, 3], keepdim=True)
        out_ln = (input - ln_mean) / torch.sqrt(ln_var + self.eps)
        out = self.rho.expand(input.shape[0], -1, -1, -1) * out_in + (1-self.rho.expand(input.shape[0], -1, -1, -1)) * out_ln
        out = out * self.gamma.expand(input.shape[0], -1, -1, -1) + self.beta.expand(input.shape[0], -1, -1, -1)
        return out


'''ResnetGenerator, 模型修改自: https://github.com/minivision-ai/photo2cartoon'''
class ResnetGenerator(nn.Module):
    def __init__(self, ngf=64, img_size=256, light=False):
        super(ResnetGenerator, self).__init__()
        self.light = light
        self.ConvBlock1 = nn.Sequential(
            nn.ReflectionPad2d(3),
            nn.Conv2d(3, ngf, kernel_size=7, stride=1, padding=0, bias=False),
            nn.InstanceNorm2d(ngf),
            nn.ReLU(True)
        )
        self.HourGlass1 = HourGlass(ngf, ngf)
        self.HourGlass2 = HourGlass(ngf, ngf)
        # Down-Sampling
        self.DownBlock1 = nn.Sequential(
            nn.ReflectionPad2d(1),
            nn.Conv2d(ngf, ngf*2, kernel_size=3, stride=2, padding=0, bias=False),
            nn.InstanceNorm2d(ngf * 2),
            nn.ReLU(True)
        )
        self.DownBlock2 = nn.Sequential(
            nn.ReflectionPad2d(1),
            nn.Conv2d(ngf*2, ngf*4, kernel_size=3, stride=2, padding=0, bias=False),
            nn.InstanceNorm2d(ngf*4),
            nn.ReLU(True)
        )
        # Encoder Bottleneck
        self.EncodeBlock1 = ResnetBlock(ngf*4)
        self.EncodeBlock2 = ResnetBlock(ngf*4)
        self.EncodeBlock3 = ResnetBlock(ngf*4)
        self.EncodeBlock4 = ResnetBlock(ngf*4)
        # Class Activation Map
        self.gap_fc = nn.Linear(ngf*4, 1)
        self.gmp_fc = nn.Linear(ngf*4, 1)
        self.conv1x1 = nn.Conv2d(ngf*8, ngf*4, kernel_size=1, stride=1)
        self.relu = nn.ReLU(True)
        # Gamma, Beta block
        if self.light:
            self.FC = nn.Sequential(nn.Linear(ngf*4, ngf*4), nn.ReLU(True), nn.Linear(ngf*4, ngf*4), nn.ReLU(True))
        else:
            self.FC = nn.Sequential(nn.Linear(img_size//4*img_size//4*ngf*4, ngf*4), nn.ReLU(True), nn.Linear(ngf*4, ngf*4), nn.ReLU(True))
        # Decoder Bottleneck
        self.DecodeBlock1 = ResnetSoftAdaLINBlock(ngf*4)
        self.DecodeBlock2 = ResnetSoftAdaLINBlock(ngf*4)
        self.DecodeBlock3 = ResnetSoftAdaLINBlock(ngf*4)
        self.DecodeBlock4 = ResnetSoftAdaLINBlock(ngf*4)
        # Up-Sampling
        self.UpBlock1 = nn.Sequential(
            nn.Upsample(scale_factor=2), 
            nn.ReflectionPad2d(1), 
            nn.Conv2d(ngf*4, ngf*2, kernel_size=3, stride=1, padding=0, bias=False),
            LIN(ngf*2),
            nn.ReLU(True)
        )
        self.UpBlock2 = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.ReflectionPad2d(1),
            nn.Conv2d(ngf*2, ngf, kernel_size=3, stride=1, padding=0, bias=False),
            LIN(ngf),
            nn.ReLU(True)
        )
        self.HourGlass3 = HourGlass(ngf, ngf)
        self.HourGlass4 = HourGlass(ngf, ngf, False)
        self.ConvBlock2 = nn.Sequential(
            nn.ReflectionPad2d(3),
            nn.Conv2d(3, 3, kernel_size=7, stride=1, padding=0, bias=False),
            nn.Tanh()
        )
    '''forward'''
    def forward(self, x):
        x = self.ConvBlock1(x)
        x = self.HourGlass1(x)
        x = self.HourGlass2(x)
        x = self.DownBlock1(x)
        x = self.DownBlock2(x)
        x = self.EncodeBlock1(x)
        content_features1 = F.adaptive_avg_pool2d(x, 1).view(x.shape[0], -1)
        x = self.EncodeBlock2(x)
        content_features2 = F.adaptive_avg_pool2d(x, 1).view(x.shape[0], -1)
        x = self.EncodeBlock3(x)
        content_features3 = F.adaptive_avg_pool2d(x, 1).view(x.shape[0], -1)
        x = self.EncodeBlock4(x)
        content_features4 = F.adaptive_avg_pool2d(x, 1).view(x.shape[0], -1)
        gap = F.adaptive_avg_pool2d(x, 1)
        gap_logit = self.gap_fc(gap.view(x.shape[0], -1))
        gap_weight = list(self.gap_fc.parameters())[0]
        gap = x * gap_weight.unsqueeze(2).unsqueeze(3)
        gmp = F.adaptive_max_pool2d(x, 1)
        gmp_logit = self.gmp_fc(gmp.view(x.shape[0], -1))
        gmp_weight = list(self.gmp_fc.parameters())[0]
        gmp = x * gmp_weight.unsqueeze(2).unsqueeze(3)
        cam_logit = torch.cat([gap_logit, gmp_logit], 1)
        x = torch.cat([gap, gmp], 1)
        x = self.relu(self.conv1x1(x))
        heatmap = torch.sum(x, dim=1, keepdim=True)
        if self.light:
            x_ = F.adaptive_avg_pool2d(x, 1)
            style_features = self.FC(x_.view(x_.shape[0], -1))
        else:
            style_features = self.FC(x.view(x.shape[0], -1))
        x = self.DecodeBlock1(x, content_features4, style_features)
        x = self.DecodeBlock2(x, content_features3, style_features)
        x = self.DecodeBlock3(x, content_features2, style_features)
        x = self.DecodeBlock4(x, content_features1, style_features)
        x = self.UpBlock1(x)
        x = self.UpBlock2(x)
        x = self.HourGlass3(x)
        x = self.HourGlass4(x)
        out = self.ConvBlock2(x)
        return out, cam_logit, heatmap


'''人脸卡通化'''
class CartoonizeFaceBeautifier(BaseBeautifier):
    def __init__(self, use_cuda=False, use_face_segmentor=True, **kwargs):
        super(CartoonizeFaceBeautifier, self).__init__(**kwargs)
        self.model_urls = {
            'transformer': 'https://github.com/CharlesPikachu/pydrawing/releases/download/checkpoints/cartoonizeface_transformer.pth',
        }
        self.use_cuda = use_cuda
        self.use_face_segmentor = use_face_segmentor
        self.face_detector = FaceDetector(use_cuda=(torch.cuda.is_available() and self.use_cuda))
        self.transformer = ResnetGenerator(ngf=32, img_size=256, light=True)
        self.transformer.load_state_dict(model_zoo.load_url(self.model_urls['transformer'])['genA2B'])
        self.transformer.eval()
        if use_face_segmentor: 
            self.face_segmentor = FaceSegmentor()
            self.face_segmentor.eval()
        if torch.cuda.is_available() and self.use_cuda: 
            self.transformer = self.transformer.cuda()
            if use_face_segmentor: self.face_segmentor = self.face_segmentor.cuda()
    '''迭代图片'''
    def iterimage(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        # 人脸提取
        face_rgb = self.face_detector(image)
        # 人脸分割
        if self.use_face_segmentor:
            face_rgb_for_seg = self.face_segmentor.preprocess(face_rgb)
            mask = self.face_segmentor(face_rgb_for_seg)
            mask = F.interpolate(mask, size=face_rgb.shape[:2][::-1], mode='bilinear', align_corners=False)
            mask = mask[0].argmax(0).cpu().numpy().astype(np.int32)
            mask = self.face_segmentor.getfacemask(mask)
        else:
            mask = np.ones(face_rgb.shape[:2]) * 255
        mask = mask[:, :, np.newaxis]
        face_rgba = np.dstack((face_rgb, mask))
        # 人脸处理
        face_rgba = cv2.resize(face_rgba, (256, 256), interpolation=cv2.INTER_AREA)
        face = face_rgba[:, :, :3].copy()
        mask = face_rgba[:, :, 3][:, :, np.newaxis].copy() / 255.
        face = (face * mask + (1 - mask) * 255) / 127.5 - 1
        face = np.transpose(face[np.newaxis, :, :, :], (0, 3, 1, 2)).astype(np.float32)
        face = torch.from_numpy(face).type(torch.FloatTensor)
        if torch.cuda.is_available() and self.use_cuda: 
            face = face.cuda()
        # 推理
        with torch.no_grad(): 
            face_cartoon = self.transformer(face)[0][0]
        # 后处理
        face_cartoon = np.transpose(face_cartoon.cpu().numpy(), (1, 2, 0))
        face_cartoon = (face_cartoon + 1) * 127.5
        face_cartoon = (face_cartoon * mask + 255 * (1 - mask)).astype(np.uint8)
        face_cartoon = cv2.cvtColor(face_cartoon, cv2.COLOR_RGB2BGR)
        # 返回
        return face_cartoon


'''
Function:
    复现论文"Combining Sketch and Tone for Pencil Drawing Production"
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import cv2
import math
import numpy as np
from PIL import Image
from scipy import signal
from ..base import BaseBeautifier
from scipy.ndimage import interpolation
from scipy.sparse.linalg import spsolve
from scipy.sparse import csr_matrix, spdiags


'''图像处理工具'''
class ImageProcessor():
    '''将像素值压缩到[0, 1]'''
    @staticmethod
    def im2double(img):
        if len(img.shape) == 2: return (img - img.min()) / (img.max() - img.min())
        else: return cv2.normalize(img.astype('float'), None, 0.0, 1.0, cv2.NORM_MINMAX)
    '''拉普拉斯分布'''
    @staticmethod
    def Laplace(x, sigma=9):
        value = (1. / sigma) * math.exp(-(256 - x) / sigma) * (256 - x)
        return value
    '''均匀分布'''
    @staticmethod
    def Uniform(x, ua=105, ub=225):
        value = (1. / (ub - ua)) * (max(x - ua, 0) - max(x - ub, 0))
        return value
    '''高斯分布'''
    @staticmethod
    def Gaussian(x, u=90, sigma=11):
        value = (1. / math.sqrt(2 * math.pi * sigma)) * math.exp(-((x - u) ** 2) / (2 * (sigma ** 2)))
        return value
    '''水平方向拼接'''
    @staticmethod
    def horizontalStitch(img, width):
        img_stitch = img.copy()
        while img_stitch.shape[1] < width:
            window_size = int(round(img.shape[1] / 4.))
            left = img[:, (img.shape[1]-window_size): img.shape[1]]
            right = img[:, :window_size]
            aleft = np.zeros((left.shape[0], window_size))
            aright = np.zeros((left.shape[0], window_size))
            for i in range(window_size):
                aleft[:, i] = left[:, i] * (1 - (i + 1.) / window_size)
                aright[:, i] = right[:, i] * (i + 1.) / window_size
            img_stitch = np.column_stack((img_stitch[:, :(img_stitch.shape[1]-window_size)], aleft+aright, img_stitch[:, window_size: img_stitch.shape[1]]))
        img_stitch = img_stitch[:, :width]
        return img_stitch
    '''垂直方向拼接'''
    @staticmethod
    def verticalStitch(img, height):
        img_stitch = img.copy()
        while img_stitch.shape[0] < height:
            window_size = int(round(img.shape[0] / 4.))
            up = img[(img.shape[0]-window_size): img.shape[0], :]
            down = img[0:window_size, :]
            aup = np.zeros((window_size, up.shape[1]))
            adown = np.zeros((window_size, up.shape[1]))
            for i in range(window_size):
                aup[i, :] = up[i, :] * (1 - (i + 1.) / window_size)
                adown[i, :] = down[i, :] * (i + 1.) / window_size
            img_stitch = np.row_stack((img_stitch[:img_stitch.shape[0]-window_size, :], aup+adown, img_stitch[window_size: img_stitch.shape[0], :]))
        img_stitch = img_stitch[:height, :]
        return img_stitch


'''复现论文"Combining Sketch and Tone for Pencil Drawing Production"'''
class PencilDrawingBeautifier(BaseBeautifier):
    def __init__(self, mode='gray', kernel_size_scale=1/40, stroke_width=1, color_depth=1, weights_color=[62, 30, 5], weights_gray=[76, 22, 2], texture_path=None, **kwargs):
        super(PencilDrawingBeautifier, self).__init__(**kwargs)
        assert mode in ['gray', 'color']
        self.rootdir = os.path.split(os.path.abspath(__file__))[0]
        self.image_processor = ImageProcessor()
        self.mode = mode
        # 铅笔笔画相关参数
        self.kernel_size_scale, self.stroke_width = kernel_size_scale, stroke_width
        # 铅笔色调相关参数
        self.weights_color, self.weights_gray, self.color_depth = weights_color, weights_gray, color_depth
        if (texture_path is None) or (not os.path.exists(texture_path)): self.texture_path = os.path.join(self.rootdir, 'textures/default.jpg')
    '''迭代图片'''
    def iterimage(self, image):
        if self.mode == 'color':
            img = Image.fromarray(image)
            img_ycbcr = img.convert('YCbCr')
            img = np.ndarray((img.size[1], img.size[0], 3), 'u1', img_ycbcr.tobytes())
            img_out = img.copy()
            img_out.flags.writeable = True
            img_out[:, :, 0] = self.__strokeGeneration(img[:, :, 0]) * self.__toneGeneration(img[:, :, 0]) * 255
            img_out = cv2.cvtColor(img_out, cv2.COLOR_YCR_CB2BGR)
        else:
            img = image
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            img_s = self.__strokeGeneration(img)
            img_t = self.__toneGeneration(img)
            img_out = img_s * img_t * 255
        return img_out
    '''铅笔笔画生成'''
    def __strokeGeneration(self, img):
        h, w = img.shape
        kernel_size = int(min(w, h) * self.kernel_size_scale)
        kernel_size += kernel_size % 2
        # 计算梯度，产生幅度
        img_double = self.image_processor.im2double(img)
        dx = np.concatenate((np.abs(img_double[:, :-1]-img_double[:, 1:]), np.zeros((h, 1))), 1)
        dy = np.concatenate((np.abs(img_double[:-1, :]-img_double[1:, :]), np.zeros((1, w))), 0)
        img_gradient = np.sqrt(np.power(dx, 2) + np.power(dy, 2))
        # 选择八个参考方向
        line_segments = np.zeros((kernel_size, kernel_size, 8))
        for i in [0, 1, 2, 7]:
            for x in range(kernel_size):
                y = round((x + 1 - kernel_size / 2) * math.tan(math.pi / 8 * i))
                y = kernel_size / 2 - y
                if y > 0 and y <= kernel_size:
                    line_segments[int(y-1), x, i] = 1
                if i == 7:
                    line_segments[:, :, 3] = np.rot90(line_segments[:, :, 7], -1)
                else:
                    line_segments[:, :, i+4] = np.rot90(line_segments[:, :, i], 1)
        # 获取参考方向的响应图
        response_maps = np.zeros((h, w, 8))
        for i in range(8):
            response_maps[:, :, i] = signal.convolve2d(img_gradient, line_segments[:, :, i], 'same')
        response_maps_maxvalueidx = response_maps.argmax(axis=-1)
        # 通过在所有方向的响应中选择最大值来进行分类
        magnitude_maps = np.zeros_like(response_maps)
        for i in range(8):
            magnitude_maps[:, :, i] = img_gradient * (response_maps_maxvalueidx == i).astype('float')
        # 线条整形
        stroke_maps = np.zeros_like(response_maps)
        for i in range(8):
            stroke_maps[:, :, i] = signal.convolve2d(magnitude_maps[:, :, i], line_segments[:, :, i], 'same')
        stroke_maps = stroke_maps.sum(axis=-1)
        stroke_maps = (stroke_maps - stroke_maps.min()) / (stroke_maps.max() - stroke_maps.min())
        stroke_maps = (1 - stroke_maps) ** self.stroke_width
        return stroke_maps
    '''铅笔色调生成'''
    def __toneGeneration(self, img, mode=None):
        height, width = img.shape
        # 直方图匹配
        img_hist_match = self.__histogramMatching(img, mode) ** self.color_depth
        # 获得纹理
        texture = cv2.imread(self.texture_path)
        texture = cv2.cvtColor(texture, cv2.COLOR_BGR2GRAY)[99: texture.shape[0]-100, 99: texture.shape[1]-100]
        ratio = 0.2 * min(img.shape[0], img.shape[1]) / float(1024)
        texture = interpolation.zoom(texture, (ratio, ratio))
        texture = self.image_processor.im2double(texture)
        texture = self.image_processor.horizontalStitch(texture, img.shape[1])
        texture = self.image_processor.verticalStitch(texture, img.shape[0])
        size = img.size
        nzmax = 2 * (size-1)
        i = np.zeros((nzmax, 1))
        j = np.zeros((nzmax, 1))
        s = np.zeros((nzmax, 1))
        for m in range(1, nzmax+1):
            i[m-1] = int(math.ceil((m + 0.1) / 2)) - 1
            j[m-1] = int(math.ceil((m - 0.1) / 2)) - 1
            s[m-1] = -2 * (m % 2) + 1
        dx = csr_matrix((s.T[0], (i.T[0], j.T[0])), shape=(size, size))
        nzmax = 2 * (size - img.shape[1])
        i = np.zeros((nzmax, 1))
        j = np.zeros((nzmax, 1))
        s = np.zeros((nzmax, 1))
        for m in range(1, nzmax+1):
            i[m-1, :] = int(math.ceil((m - 1 + 0.1) / 2) + img.shape[1] * (m % 2)) - 1
            j[m-1, :] = math.ceil((m - 0.1) / 2) - 1
            s[m-1, :] = -2 * (m % 2) + 1
        dy = csr_matrix((s.T[0], (i.T[0], j.T[0])), shape=(size, size))
        texture_sparse = spdiags(np.log(np.reshape(texture.T, (1, texture.size), order="f") + 0.01), 0, size, size)
        img_hist_match1d = np.log(np.reshape(img_hist_match.T, (1, img_hist_match.size), order="f").T + 0.01)
        nat = texture_sparse.T.dot(img_hist_match1d)
        a = np.dot(texture_sparse.T, texture_sparse)
        b = dx.T.dot(dx)
        c = dy.T.dot(dy)
        mat = a + 0.2 * (b + c)
        beta1d = spsolve(mat, nat)
        beta = np.reshape(beta1d, (img.shape[0], img.shape[1]), order="c")
        tone = texture ** beta
        tone = (tone - tone.min()) / (tone.max() - tone.min())
        return tone
    '''直方图匹配'''
    def __histogramMatching(self, img, mode=None):
        weights = self.weights_color if mode == 'color' else self.weights_gray
        # 图像
        histogram_img = cv2.calcHist([img], [0], None, [256], [0, 256])
        histogram_img.resize(histogram_img.size)
        histogram_img /= histogram_img.sum()
        histogram_img_cdf = np.cumsum(histogram_img)
        # 自然图像
        histogram_natural = np.zeros_like(histogram_img)
        for x in range(256):
            histogram_natural[x] = weights[0] * self.image_processor.Laplace(x) + weights[1] * self.image_processor.Uniform(x) + weights[2] * self.image_processor.Gaussian(x)
        histogram_natural /= histogram_natural.sum()
        histogram_natural_cdf = np.cumsum(histogram_natural)
        # 做直方图匹配
        img_hist_match = np.zeros_like(img)
        for x in range(img.shape[0]):
            for y in range(img.shape[1]):
                value = histogram_img_cdf[img[x, y]]
                img_hist_match[x, y] = (np.abs(histogram_natural_cdf-value)).argmin()
        img_hist_match = np.true_divide(img_hist_match, 255)
        return img_hist_match